import _ from 'lodash';
import {observable, IObservableArray} from 'mobx';
import { PageItems } from '../tool';
import {Field, ArrFields} from './uqMan';
import {Entity} from './entity';
import { QueryQueryCaller, QueryPageCaller } from './caller';

export type QueryPageApi = (name:string, pageStart:any, pageSize:number, params:any) => Promise<string>;

export class QueryPager<T extends any> extends PageItems<T> {
    private query: Query;
    constructor(query: Query, pageSize?: number, firstSize?: number, itemObservable?:boolean) {
        super(itemObservable);
        this.query = query;
        if (pageSize !== undefined) this.pageSize = pageSize;
        if (firstSize !== undefined) this.firstSize = firstSize;
    }

	setReverse() {
		this.appendPosition = 'head';
	}

    protected async onLoad() {
        let {schema} = this.query;
        if (schema === undefined) await this.query.loadSchema();
    }

    protected async loadResults(param:any, pageStart:number, pageSize:number):Promise<{[name:string]:any[]}> {
		let ret = await this.query.page(param, pageStart, pageSize);
		return ret;
	}
	private get$Page() {
        let {schema} = this.query;
        if (schema === undefined) return;
        let $page = (schema.returns as any[]).find(v => v.name === '$page');
		if ($page === undefined) return;
		return $page;
	}
	private getPageStart(item:T) {
		let $page = this.get$Page();
        let {order} = $page;
        if (order === undefined) return;
		if (item === undefined) return;
		let field = $page.fields[0];
		if (!field) return;

		let start = item[field.name];
		if (start === null) return;
		if (start === undefined) return;
		if (typeof start === 'object') {
			return start.id;
		}
		return start;
	}
    protected setPageStart(item:T) {
		this.pageStart = this.getPageStart(item);
	}
	findItem(id:any):T {
		let $page = this.get$Page();
		if (!$page) return;
        let {fields} = $page;
		let field = fields[0];
		if (!field) return;
		let fn = field.name;

		let newId = id;
		if (newId === undefined || newId === null) return;
		if (typeof newId === 'object') newId = newId.id;
		let oldItem = this._items.find(v => {
			let oldId = v[fn];
			if (oldId === undefined || oldId === null) return false;
			if (typeof oldId === 'object') oldId = oldId.id;
			return oldId = newId;
		});
		return oldItem;
	}
	async refreshItems(item:T) {
		let $page = this.get$Page();
		if (!$page) return;
        let {fields} = $page;
		let index = this._items.indexOf(item);
		if (index < 0) return;
		let startIndex:number;
		if (this.appendPosition === 'tail') {
			startIndex = index - 1;
		}
		else {
			startIndex = index + 1;
		}
		let field = fields[0];
		if (!field) return;
		let pageStart = this.getPageStart(this._items[startIndex]);
		let pageSize = 1;
        let ret = await this.load(
			this.param, 
			pageStart,
			pageSize);
		let len = ret.length;
		if (len === 0) {
			this._items.splice(index, 1);
			return;
		}
		let fn = field.name;
		for (let i=0; i<len; i++) {
			let newItem = ret[i];
			if (!newItem) continue;
			let newId = newItem[fn];
			if (newId === undefined || newId === null) continue;
			if (typeof newId === 'object') newId = newId.id;
			let oldItem = this._items.find(v => {
				let oldId = v[fn];
				if (oldId === undefined || oldId === null) return false;
				if (typeof oldId === 'object') oldId = oldId.id;
				return oldId = newId;
			});
			if (oldItem) {
				_.merge(oldItem, newItem);
			}
		}
	}
}

export class Query extends Entity {
    get typeName(): string { return 'query';}
    private pageStart: any;
    private pageSize:number;
    private params:any;
    private more: boolean;
    private startField: Field;
    @observable list:IObservableArray; // = observable.array([], {deep: false});
    returns: ArrFields[];
    isPaged: boolean;

    setSchema(schema:any) {
        super.setSchema(schema);
        let {returns} = schema;
        //this.returns = returns;
        this.isPaged = returns && (returns as any[]).find(v => v.name === '$page') !== undefined;
    }

    resetPage(size:number, params:any) {
        this.pageStart = undefined;
        this.pageSize = size;
        this.params = params;
        this.more = false;
        this.list = undefined;
    }
    get hasMore() {return this.more;}
    async loadPage():Promise<void> {
        if (this.pageSize === undefined) {
            throw new Error('call resetPage(size:number, params:any) first');
        }
        let pageStart:any;
        if (this.pageStart !== undefined) {
            switch (this.startField.type) {
                default: pageStart = this.pageStart; break;
                case 'date':
                case 'time':
                case 'datetime': pageStart = (this.pageStart as Date).getTime(); break;
            }
        }
		let ret = await this.page(this.params, pageStart, this.pageSize+1);
		let page = ret.$page;
        /*
        await this.loadSchema();
        let res = await this.tvApi.page(this.name, pageStart, this.pageSize+1, this.params);
        let data = await this.unpackReturns(res);
        let page = data['$page'] as any[];
        */
        this.list = observable.array([], {deep: false});
        if (page !== undefined) {
            if (page.length > this.pageSize) {
                this.more = true;
                page.pop();
                let ret = this.returns.find(r => r.name === '$page');
                this.startField = ret.fields[0];
                this.pageStart = page[page.length-1][this.startField.name];
            }
            else {
                this.more = false;
            }
            this.list.push(...page);
        }
        //this.loaded = true;
    }

    protected pageCaller(params: any, showWaiting: boolean = true): QueryPageCaller {
        return new QueryPageCaller(this, params, showWaiting);
    }

    async page(params:any, pageStart:any, pageSize:number, showWaiting: boolean = true):Promise<{[name:string]:any[]}> {
        /*
        await this.loadSchema();
        let res = await this.uqApi.page(this.name, pageStart, pageSize+1, this.buildParams(params));
        */
        let p = {pageStart:pageStart, pageSize:pageSize, params:params};
        let res = await this.pageCaller(p, showWaiting).request();
        //let data = this.unpackReturns(res);
        //return data.$page;// as any[];
        return res;
    }
    protected queryCaller(params: any, showWaiting: boolean = true): QueryQueryCaller {
        return new QueryQueryCaller(this, params, showWaiting);
    }
    async query(params:any, showWaiting:boolean = true):Promise<any> {
        /*
        await this.loadSchema();
        let res = await this.uqApi.query(this.name, this.buildParams(params));
        */
        let res = await this.queryCaller(params, showWaiting).request();
        //let data = this.unpackReturns(res);
        //return data;
        return res;
    }
    async table(params:any, showWaiting:boolean = true): Promise<any[]> {
        let ret = await this.query(params, showWaiting);
        for (let i in ret) {
            return ret[i];
        }
    }
    async obj(params:any, showWaiting:boolean = true):Promise<any> {
        let ret = await this.table(params, showWaiting);
        if (ret.length > 0) return ret[0];
    }
    async scalar(params:any, showWaiting:boolean = true):Promise<any> {
        let ret = await this.obj(params, showWaiting);
        for (let i in ret) return ret[i];
    }
}
